
#include "Burbulator.h"
#include "Resources.h"
#include "Fluid.h"
#include "WaterSurface.h"
#include "../utils/Random.h"

const float MAX_AGE = 4.0f;
const float MAX_SIZE = 15.0f;
const float MIN_TEMPERATURE = 80.0f;
const float MIN_TEMPERATURE2 = 96.0f;
const float INI_SIZE = 0.4f;
const float TEMPERATURE_SIZE = 0.05f;
const float TEMPERATURE_SIZE2 = 0.05f;
const float RISE_AMOUNT = 2.0f;
const int BUBBLES_PER_SECOND = 600.0f;

void CBurbulator::update( float deltaTime, CWaterSurface& surface )
{
	CFluid& fluid = *mResources->mFluid;
	const float scale = (fluid.getSize()/2) / R_FLUID_RADIUS;
	const float step = R_FLUID_RADIUS / (fluid.getSize()/2);

	float invDeltaTime = 1.0f / (0.0001f+deltaTime);

	//
	// spawn new bubbles
	
	if( mBubbles.size() < 2000 ) {
		int bubs = BUBBLES_PER_SECOND * deltaTime + 1; // TBD: more accurate
		for( int i = 0; i < bubs; ++i ) {
			float x = random::randfs( R_FLUID_RADIUS );
			float y = random::randfs( R_FLUID_RADIUS );
			if( x*x+y*y >= R_FLUID_RADIUS2 )
				continue;
			float z = random::randf( R_FLUID_HEIGHT * 0.90f ) + R_FLUID_HEIGHT * 0.05f;
			int ix = (int)(x*scale) + fluid.getSize()/2;
			int iy = (int)(y*scale) + fluid.getSize()/2;
			int iz = (int)(z*scale);
			int idx = fluid.getIndex(ix,iy,iz);
			float temp = fluid.getTemperature()[idx];
			// spawn bubble
			if( temp > MIN_TEMPERATURE ) {
				SBubble& b = mBubbles.add();
				b.mPosition.x = x;
				b.mPosition.y = y;
				b.mPosition.z = z;
				b.mSize = INI_SIZE + (temp-MIN_TEMPERATURE)*TEMPERATURE_SIZE;
				b.mAge = 0.0f;
				b.mWasAtSurface = false;
				// disturb fluid
				fluid.getVelocityX0()[idx] += random::randfs( 0.02f ) * invDeltaTime;
				fluid.getVelocityY0()[idx] += random::randfs( 0.02f ) * invDeltaTime;
				fluid.getVelocityZ0()[idx] += random::randf( 0.01f ) * invDeltaTime;
			}
		}
	}

	//
	// update all bubbles - move by fluid velocity, grow by fluid temperature,
	// die if exceeded max size/age

	TBubblePool::iterator it;
	for( it = mBubbles.begin(); it != mBubbles.end(); /**/ ) {
		SBubble& b = *it;
		// get fluid params
		fluid.interpolatePrepare( b.mPosition.x * scale, b.mPosition.y * scale, b.mPosition.z * scale );
		float velx = fluid.interpolate( fluid.getVelocityX() ) * deltaTime;
		float vely = fluid.interpolate( fluid.getVelocityY() ) * deltaTime;
		float velz = fluid.interpolate( fluid.getVelocityZ() ) * deltaTime;
		float temp = fluid.interpolate( fluid.getTemperature() );
		// move bubble
		D3DXVECTOR3 bubvel;
		bubvel.x = velx * R_FLUID_RADIUS * 6.0f;
		bubvel.y = vely * R_FLUID_RADIUS * 6.0f;
		bubvel.z = velz * R_FLUID_RADIUS;
		float riseZ = b.mSize*b.mSize*b.mSize*3.14f*RISE_AMOUNT * deltaTime;
		bubvel.z += riseZ;
		b.mPosition += bubvel;
		// grow bubble
		if( temp > MIN_TEMPERATURE2 && b.mWasAtSurface )
			b.mSize += (temp-MIN_TEMPERATURE2)*TEMPERATURE_SIZE2;
		// when bubble reaches the surface, inject force into fluid downwards
		float overz = b.mPosition.z - R_FLUID_HEIGHT;
		if( overz > 0 ) {
			D3DXVECTOR3 vel;
			vel.x = random::randfs( 0.007f ) * invDeltaTime;
			vel.y = random::randfs( 0.007f ) * invDeltaTime;
			vel.z = -random::randf( 0.012f ) * invDeltaTime;
			if( !b.mWasAtSurface ) {
				int idx = fluid.getInterpolate().idx[rand()&7];
				fluid.getVelocityX0()[idx] += vel.x;
				fluid.getVelocityY0()[idx] += vel.y;
				fluid.getVelocityZ0()[idx] += vel.z;
				b.mWasAtSurface = true;

				// ==>
			}
			bubvel += vel;
			b.mPosition.z = R_FLUID_HEIGHT;
		}
		// feed into surface
		surface.feedVelocity( b.mPosition.x, b.mPosition.y, bubvel, 1.0f );
		// limit horizontal position
		float dist2 = b.mPosition.x * b.mPosition.x + b.mPosition.y * b.mPosition.y;
		float dist = sqrtf( dist2 );
		if( dist + b.mSize > R_FLUID_RADIUS ) {
			float f = (R_FLUID_RADIUS-b.mSize) / dist;
			b.mPosition.x *= f;
			b.mPosition.y *= f;
		}
		// limit bottom
		if( b.mPosition.z < R_FLUID_HEIGHT * 0.05f )
			b.mPosition.z = R_FLUID_HEIGHT * 0.05f;
		// manage age
		b.mAge += deltaTime;
		
		// see if should die
		if( b.mAge > MAX_AGE || b.mSize > MAX_SIZE || temp < MIN_TEMPERATURE * 0.7f ) {
			it = mBubbles.erase( it );
		} else {
			++it;
		}
	}
}

